/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <iostream.h>
#include <iomanip.h>

//#include "14bit-registers.h"
//#include "14bit-instructions.h"
#include "14bit-processors.h"

#include <g++/string>
#include "stimuli.h"


//-------------------------------------------------------------------
//
// 
//    create
//
//  The purpose of this member function is to 'create' those things
// that are unique to the 14-bit core processors.

void _14bit_processor :: create (void)
{

  cout << "_14bit_processor create, type = " << isa() << '\n';

  pic_processor::create();

}

//
// create_symbols
//
//  Create symbols for a generic 14-bit core. This allows symbolic
// access to the pic. (e.g. It makes it possible to access the 
// status register by name instead of by its address.)
//

void _14bit_processor::create_symbols (void)
{

  cout << "14bit create symbols\n";

}


//-------------------------------------------------------------------
void _14bit_processor::
interrupt (void)
{
  
  bp.clear_interrupt();

  stack.push(pc.value);
  intcon->clear_gie();

  pc.jump(INTERRUPT_VECTOR);

}

//-------------------------------------------------------------------
void _14bit_processor::por(void)
{
  pic_processor::por();
}

//-------------------------------------------------------------------
void _14bit_processor::option_new_bits_6_7(unsigned int bits)
{

  //portb.rbpu_intedg_update(bits);
  cout << "14bit, option bits 6 and/or 7 changed\n";

}


//-------------------------------------------------------------------
void PORTB::rbpu_intedg_update(unsigned int new_configuration)
{
  int i;

  if((new_configuration ^ rbpu) & rbpu_MASK )
    {
      rbpu = new_configuration & rbpu_MASK;
      for(i=0; i<8; i++)
	( (IO_bi_directional_pu *)pins[i]) ->pull_up_resistor->drive = rbpu;
    }

  if((new_configuration ^ intedg) & intedg_MASK )
    {
      intedg = new_configuration & intedg_MASK;
    }
}


unsigned int PORTB::get(void)
{
  unsigned int old_value;

  old_value = value;

  //  cout << "PORTB::get()\n";
  IOPORT::get();

  int diff = old_value ^ value; // The difference between old and new

  // If portb bit 0 changed states, check to see if an interrupt should occur
  if( diff & 1)
    {
      // If the option register is selecting 'interrupt on rising edge' AND
      // the old_value was low (and hence the new high) then set  INTF
      //                    OR
      // If the option register is selecting 'interrupt on falling edge' AND
      // the old_value was high (and hence the new low) then set  INTF

      // These two statements can be combined into an exclusive or:

      if( (intedg &  intedg_MASK) ^ (old_value & 1))
	cpu14->intcon->set_intf();
    }


  // If any of the upper 4 bits of port b changed states then set RBIF
  if(diff & 0xf0)
    cpu14->intcon->set_rbif();
}

void PORTB::setbit(unsigned int bit_number, bool new_value)
{
  unsigned int old_value = value;

  //cout << "PORTB::setbit() bit " << bit_number << " to " << new_value << '\n';

  IOPORT::setbit( bit_number,  new_value);

  int diff = old_value ^ value; // The difference between old and new

  // If portb bit 0 changed states, check to see if an interrupt should occur
  if( diff & 1)
    {
      // If the option register is selecting 'interrupt on rising edge' AND
      // the old_value was low (and hence the new high) then set  INTF
      //                    OR
      // If the option register is selecting 'interrupt on falling edge' AND
      // the old_value was high (and hence the new low) then set  INTF

      // These two statements can be combined into an exclusive or:
      //cout << "PORTB::setbit() - bit changed states\n";
      if( (intedg &  intedg_MASK) ^ (old_value & 1))
	cpu14->intcon->set_intf();
    }


  // If any of the upper 4 bits of port b changed states then set RBIF
  if(diff & 0xf0)
    cpu14->intcon->set_rbif();
}


void PORTA::setbit(unsigned int bit_number, bool new_value)
{

  unsigned int old_value = value;

  //cout << "PORTA::setbit() bit " << bit_number << " to " << new_value << '\n';

  IOPORT::setbit( bit_number,  new_value);

  int diff = old_value ^ value; // The difference between old and new

  // If porta bit 4 changed states, check to see if tmr0 should increment
  if( diff & 0x10)
    {
      //cout << "bit 4 changed\n";

      if(cpu14->option_reg.get_t0cs())
	{
	  //  cout << "tmr 0 external clock, porta new value " << value << " t0se "<<cpu14->option_reg.get_t0se()<< '\n';
	if( ((value & 0x10) == 0) ^ (cpu14->option_reg.get_t0se() == 0))
	  cpu14->tmr0.increment();
	}
    }

}
