/*
   Copyright (C) 1998 Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include <stdio.h>
#include <iostream.h>
#include <iomanip.h>


#include "14bit-processors.h"

#include <g++/string>
#include "stimuli.h"

#define PCLATH_MASK              0x1f


pic_processor *temp_cpu;

//--------------------------------------------------
// Member functions for the file_register base class
//--------------------------------------------------
//

file_register::file_register(void)
{
  name_str1 = NULL;
  new_name("file_register");
}

//--------------------------------------------------
// get
//
//  Return the contents of the file register. If a
// 'break on read' break point is set, then set the
// global break point flag.
/*
unsigned int file_register::get(void)
{
  trace.register_read(address,value);

  return(value);
}
*/
//--------------------------------------------------
//
//  Update the contents of the file register. If a
// 'break on write' break point is set and this write
// meets the conditions for the break point then set the
// global break point flag.
/*
void file_register::put(unsigned int new_value)
{

  value = new_value;
  trace.register_write(address,value);

}
*/
void file_register::new_name(char *s)
{
  if(name_str1)
    delete name_str1;

  name_str1 = new char[strlen(s)];
  strcpy(name_str1,s);
}

//--------------------------------------------------
// member functions for the invalid_file_register class
//--------------------------------------------------
void invalid_file_register::put(unsigned int new_value)
{
  cout << "attempt write to invalid file register at address " << address<< '\n';

  bp.halt();
  trace.register_write(address,value);

  return;
}

unsigned int invalid_file_register::get(void)
{
  cout << "attempt read from invalid fr\n";

  if(break_point)
    bp.check_invalid_fr_break(this);

  trace.register_read(address,value);

  return(0);
}



//
//--------------------------------------------------
// member functions for the FSR class
//--------------------------------------------------
//
void  FSR::put(unsigned int new_value)
{
  value = new_value;
  trace.register_write(address,value);

  /* The 12-bit core selects the register page using the fsr */
  if(cpu->base_isa() == _12BIT_PROCESSOR_)
    {
      cpu->register_bank = &cpu->registers[ value & register_page_bits ];
    }


}

//
//--------------------------------------------------
// member functions for the Status_register class
//--------------------------------------------------
//

//--------------------------------------------------

Status_register::Status_register(void)
{
  break_point = 0;
  break_on_z =0;
  address = 3;
  new_name("status");
}
//--------------------------------------------------
// put

void inline Status_register::put(unsigned int new_value)
{
  value = new_value; // & STATUS_WRITABLE_BITS;

  if(cpu->base_isa() == _14BIT_PROCESSOR_)
    {
      cpu->register_bank = &cpu->registers[(value & RP_MASK) << 2];
    }

  trace.register_write(address,value);
}


//--------------------------------------------------
// get
//unsigned int Status_register::get(void)

//--------------------------------------------------
// put_Z

//void Status_register::put_Z(unsigned int new_z)

//--------------------------------------------------
// get_Z
//unsigned int Status_register::get_Z(void)
//--------------------------------------------------
// put_C

//void Status_register::put_C(unsigned int new_c)

//--------------------------------------------------
// get_C
//unsigned int Status_register::get_C(void)

//--------------------------------------------------
// put_Z_C_DC

//--------------------------------------------------
// member functions for the INDF class
//--------------------------------------------------
INDF::INDF(void)
{
  break_point = 0;

  new_name("indf");
}

void INDF::put(unsigned int new_value)
{

  trace.register_write(address,value);

  if(cpu->fsr.value)
    cpu->registers[cpu->fsr.value]->put(new_value);

}

unsigned int INDF::get(void)
{

  trace.register_read(address,value);

  if(cpu->fsr.value)
    return(cpu->registers[cpu->fsr.value]->get());
  else
    return(0);   // avoid infinite loop if fsr points to the indf
}

//--------------------------------------------------
// member functions for the TMR0 base class
//--------------------------------------------------
TMR0::TMR0(void)
{
  break_point = 0;
  value=0;
  synchronized_cycle=0;
  prescale=1;
  new_name("tmr0");
}

void TMR0::start(void)
{

  value = 0;
  cout << "TMRO::start\n";

  synchronized_cycle = cpu->cycles.value;    // + 2 ???

  prescale = 1 << (cpu->option_reg.get_psa() ? 0 : (1+cpu->option_reg.get_prescale()));

  last_cycle = synchronized_cycle - value * prescale;
  future_cycle = last_cycle + 256 * prescale;

  cout << "TMR0::start   last_cycle = " << hex << last_cycle << " future_cycle = " << future_cycle << '\n';
  cpu->cycles.set_break(future_cycle, this);
  cpu->cycles.dump_breakpoints();

}


// %%%FIX ME%%% 
void TMR0::increment(void)
{
  //  cout << "TMR0 increment because of external clock ";

  if(--prescale_counter == 0)
    {
      prescale_counter = prescale;
      if(++value == 256)
	{
	  value = 0;
	  if(cpu->base_isa() == _14BIT_PROCESSOR_)
	    {
	      cpu14->intcon->set_t0if();
	    }
	}
      trace.register_write(address,value);
    }
  //  cout << value << '\n';
}

void TMR0::put(unsigned int new_value)
{
  if(cpu->option_reg.get_t0cs())
    {
      cout << "TMR0::put external clock...\n";
    }


  // Note, anytime something is written to TMR0, the prescaler, if it's
  // assigned to tmr0, is also cleared. This is implicitly handled by
  // saving the value of cpu's cycle counter and associating that value
  // with the tmr rollover.

  value = new_value & 0xff;

  prescale = 1 << (cpu->option_reg.get_psa() ? 0 : (1+cpu->option_reg.get_prescale()));
  prescale_counter = prescale;

  synchronized_cycle = cpu->cycles.value + 2;

  last_cycle = synchronized_cycle - value * prescale;

  unsigned int fc = last_cycle + 256 * prescale;

  if(future_cycle)
    cpu->cycles.reassign_break(future_cycle, fc, this);
  else
    cpu->cycles.set_break(fc, this);

  future_cycle = fc;

  trace.register_write(address,value);

}

unsigned int TMR0::get_value(void)
{
  // If the TMR0 is being read immediately after being written, then
  // it hasn't had enough time to synchronize with the PIC's clock.
  if(cpu->cycles.value <= synchronized_cycle)
    return value;

  if(cpu->option_reg.get_t0cs())
    return(value);

  int new_value = (cpu->cycles.value - last_cycle)/ prescale;

  if (new_value > 255)
    {
      cout << "TMR0: bug TMR0 is larger than 255...\n";
      return 0;
    }

  value = new_value;
  return(value);
  
}

unsigned int TMR0::get(void)
{
  value = get_value();
  trace.register_read(address, value);
}
void TMR0::new_prescale(void)
{
  cout << "tmr0 new_prescale\n";

  if(cpu->option_reg.get_t0cs())
    {
      cout << "external clock...\n";

      prescale = 1 << (cpu->option_reg.get_psa() ? 0 : (1+cpu->option_reg.get_prescale()));
      prescale_counter = prescale;
    }
  else
    {
      // Get the current value of TMR0
      int new_value = (cpu->cycles.value - last_cycle)/prescale;
      cout << "cycles " << cpu->cycles.value  << " old prescale " << prescale;

      prescale = 1 << (cpu->option_reg.get_psa() ? 0 : (1+cpu->option_reg.get_prescale()));

      cout << " new prescale " << prescale;

      // Now compute the 'last_cycle' as though if TMR0 had been running on the 
      // new prescale all along. Recall, 'last_cycle' records the value of the cpu's
      // cycle counter when tmr0 last rolled over.

      last_cycle = cpu->cycles.value - new_value * prescale;
      cout << " effective last_cycle " << last_cycle << '\n';

      if(cpu->cycles.value <= synchronized_cycle)
	last_cycle += (synchronized_cycle - cpu->cycles.value);

      unsigned int fc = last_cycle + 256 * prescale;
      cout << "moving break from " << future_cycle << " to " << fc << '\n';

      cpu->cycles.reassign_break(future_cycle, fc, this);

    }
}

void TMR0::new_clock_source(void)
{

  cout << "TMR0:new_clock_source changed to the ";
  if(cpu->option_reg.get_t0cs())
    {
      cout << "external\n";
      //      cpu->cycles.
    }
  else
    {
      cout << "internal\n";
      put(value);    // let TMR0::put() set a cycle counter break point
    }
}

// TMR0 callback is called when the cycle counter hits the break point that
// was set in TMR0::put. The cycle counter will clear the break point, so
// we don't need to worry about it. At this point, TMR0 is rolling over.

void TMR0::callback(void)
{

  //cout<<"TMR0 rollover: " << hex << cpu->cycles.value << '\n';

  // If tmr0 is being clocked by the external clock, then at some point
  // the simulate code must have switched from the internal clock to
  // external clock. The cycle break point was still set, so just ignore it.
  if(cpu->option_reg.get_t0cs())
    {
      future_cycle = 0;  // indicates that tmr0 no longer has a break point
      return;
    }

  value = 0;
  synchronized_cycle = cpu->cycles.value;
  last_cycle = synchronized_cycle;
  future_cycle = last_cycle + 256*prescale;
  cpu->cycles.set_break(future_cycle, this);
  if(cpu->base_isa() == _14BIT_PROCESSOR_)
    {
      cpu14->intcon->set_t0if();
    }


}
//--------------------------------------------------
// member functions for the OPTION base class
//--------------------------------------------------
OPTION_REG::OPTION_REG(void)
{
  break_point = 0;
  por_value = 0xff;
  wdtr_value = 0xff;
  new_name("option");
}

void OPTION_REG::put(unsigned int new_value)
{

  unsigned int old_value = value;
  value = new_value;

  // First, check the tmr0 clock source bit to see if we are  changing from
  // internal to external (or vice versa) clocks.
  if( (value ^ old_value) & T0CS)
    cpu->tmr0.new_clock_source();

  // %%%FIX ME%%% - can changing the state of TOSE cause the timer to
  // increment if tmr0 is being clocked by an external clock?

  // Now check the rest of the tmr0 bits.
  if( (value ^ old_value) & (T0SE | PSA | PS2 | PS1 | PS0))
    cpu->tmr0.new_prescale();

  if( (value ^ old_value) & (BIT6 | BIT7))
    cpu->option_new_bits_6_7(value & (BIT6 | BIT7));

  if(cpu->base_isa() == _14BIT_PROCESSOR_)
    trace.register_write(address,value);
  else
    trace.write_OPTION(value);

}



//--------------------------------------------------
// member functions for the INTCON base class
//--------------------------------------------------
INTCON::INTCON(void)
{
  break_point = 0;
  new_name("intcon");
}

void INTCON::set_T0IF(void)
{

  value |= T0IF;

  trace.register_write(address,value);

  if (value & (GIE | T0IE))
  {
    trace.interrupt();
  }
}


//--------------------------------------------------
// member functions for the PCL base class
//--------------------------------------------------
PCL::PCL(void)
{
  break_point = 0;
  new_name("pcl");
}

// %%% FIX ME %%% breaks are different
void PCL::put(unsigned int new_value)
{
  if(break_point) 
  {
    if(bp.check_write_break(this))
      cpu->pc.computed_goto(new_value);
  }
  else
    cpu->pc.computed_goto(new_value);

  trace.register_write(address,value);
}

unsigned int PCL::get(void)
{
  return((value+1) & 0xff);
}
//--------------------------------------------------
// member functions for the PCLATH base class
//--------------------------------------------------

PCLATH::PCLATH(void)
{
  break_point = 0;
  new_name("pclath");
}

void PCLATH::put(unsigned int new_value)
{
  if(break_point) 
    bp.check_write_break(this);

  value = new_value & PCLATH_MASK;

  trace.register_write(address,value);
}

unsigned int PCLATH::get(void)
{
  if(break_point) 
    bp.check_read_break(this);

  trace.register_read(address,value);
  return(value & PCLATH_MASK);
}

//--------------------------------------------------
// member functions for the Program_Counter base class
//--------------------------------------------------
//

//--------------------------------------------------

Program_Counter::Program_Counter(void)
{

  value = 0;

}

//--------------------------------------------------
// increment - update the program counter. All non-branching instructions pass through here.
//   

void Program_Counter::increment(void)
{

  value = (value + 1) & memory_size_mask;
  trace.program_counter(value);

  //  cout << "     pc increment  "  << value << '\n';

  // Update pcl. Note that we don't want to pcl.put() because that 
  // will trigger a break point if there's one set on pcl. (A read/write
  // break point on pcl should not be triggered by advancing the program
  // counter).

  cpu->pcl.value = value & 0xff;
  cpu->cycles.increment();
}

//--------------------------------------------------
// skip - Does the same thing that increment does, except that it records the operation
// in the trace buffer as a 'skip' instead of a 'pc update'.
//   

void Program_Counter::skip(void)
{

  value = (value + 1) & memory_size_mask;
  trace.pc_skip(value);


  // Update pcl. Note that we don't want to pcl.put() because that 
  // will trigger a break point if there's one set on pcl. (A read/write
  // break point on pcl should not be triggered by advancing the program
  // counter).

  cpu->pcl.value = value & 0xff;
  cpu->cycles.increment();
}

//--------------------------------------------------
// jump - update the program counter. All branching instructions except computed gotos
//        and returns go through here.

void Program_Counter::jump(unsigned int new_address)
{

  // Use the new_address and the cached pclath (or page select bits for 12 bit cores)
  // to generate the destination address:

  value = (new_address | cpu->get_pclath_branching_jump() ) & memory_size_mask;

  trace.program_counter(value);

  // cout << "pc.jump  " << value << '\n';

  cpu->pcl.value = value & 0xff;    // see Update pcl comment in Program_Counter::increment()
  cpu->cycles.increment();
  cpu->cycles.increment();
}

//--------------------------------------------------
// computed_goto - update the program counter. Anytime the pcl register is written to
//                 by the source code we'll pass through here.
//

void Program_Counter::computed_goto(unsigned int new_address)
{

  // Use the new_address and the cached pclath (or page select bits for 12 bit cores)
  // to generate the destination address:

  value = (new_address | cpu->get_pclath_branching_modpcl() ) & memory_size_mask;

  trace.program_counter(value);

  cpu->pcl.value = value & 0xff;    // see Update pcl comment in Program_Counter::increment()

  // The instruction modifying the PCL will also increment the program counter. So, pre-compensate
  // the increment with a decrement:
  value--;
  cpu->cycles.increment();
}

//--------------------------------------------------
// new_address - write a new value to the program counter. All returns pass through here.
//

void Program_Counter::new_address(unsigned int new_value)
{
  value = new_value & memory_size_mask;
  trace.program_counter(value);

  cpu->pcl.value = value & 0xff;    // see Update pcl comment in Program_Counter::increment()
  cpu->cycles.increment();
  cpu->cycles.increment();
}

//--------------------------------------------------
// get_next - get the next address that is just pass the current one
//            (used by 'call' to obtain the return address)

unsigned int Program_Counter::get_next(void)
{

  return( (value+1) & memory_size_mask);

}


//--------------------------------------------------
Stack::Stack(void)
{

  stack_warnings_flag = 1;   // Display over/under flow stack warnings
  stack_mask = 7;            // Assume a 14 bit core.
  pointer = 0;

  for(int i=0; i<8; i++)
    contents[i] = 0;

}

//
// Stack::push
//
// push the passed address onto the stack by storing it at the current
// 

void Stack::push(unsigned int address)
{

  // Write the address at the current point location. Note that the '& stack_mask'
  // implicitly handles the stack wrap around.

  contents[pointer & stack_mask] = address;

  // If the stack pointer is too big, then the stack has definitely over flowed.
  // However, some pic programs take advantage of this 'feature', so provide a means
  // for them to ignore the warnings.

  if( (pointer++ >= stack_mask) && stack_warnings_flag)
    cout << "%%% stack overflow";

}

//
// Stack::pop
//

unsigned int Stack::pop(void)
{

  // First decrement the stack pointer.

  if((--pointer < 0)  && stack_warnings_flag)
    cout << "%%% stack underflow";

  
  return(contents[pointer & stack_mask]);
}

//--------------------------------------------------
// member functions for the Cycle_Counter class
//--------------------------------------------------
//
// %%% FIX ME %%% the cycle counter is only good for 31 bits,
// or 2 billion cycles. That sounds like a lot, but in reality
// it's less than seven minutes for a 20Mhz pic. Expect the
// cycle counter logic to change in the future to accomodate
// a larger variable.

//--------------------------------------------------

void Cycle_Counter::preset(unsigned int new_value)
{
  value = new_value;
  trace.cycle_counter(value);
}

//--------------------------------------------------
// set_break
// set a cycle counter break point. Return 1 if successful.
//
//  The break points are stored in a singly-linked-list sorted by the
// order in which they will occur. When this routine is called, the
// value of 'future_cycle' is compared against the values in the
// 'active' list.

bool Cycle_Counter::set_break(unsigned int future_cycle, BreakCallBack *f=NULL, unsigned int bpn=MAX_BREAKPOINTS)
{

  Cycle_Counter_breakpoint_list  *l1 = &active, *l2;

  if(inactive.next == NULL)
    {
      cout << " too many breaks are set on the cycle counter \n";
      exit(1);
      return 0;
    }
  else if(future_cycle <= value)
    {
      cout << "Cycle break point was ignored because cycle " << future_cycle << " has already gone by\n";
      return 0;
    }
  else
    {

      // place the future cycle at which we intend to break into the
      // sorted break list

      bool break_set = 0;

      while( (l1->next) && !break_set)
	{

	  // If the next break point is at a cycle greater than the
	  // one we wish to set, then we found the insertion point.
	  // Otherwise 
	  if(l1->next->break_value > future_cycle)
	    break_set = 1;
	  else
	    l1 = l1->next;

	}

      l2 = l1->next;
      l1->next = inactive.next;
      inactive.next = l1->next->next;
      l1->next->next = l2;
      l1->next->break_value = future_cycle;
      l1->next->f = f;
      l1->next->breakpoint_number = bpn;
      //cout << "cycle break " << future_cycle << " bpn " << bpn << '\n';

    }

  break_on_this = active.next->break_value;

  return 1;
}

// reassign_break
//   change the cycle of an existing break point.
//
//  This is only called by the internal peripherals and not (directly) by the user. It's
// purpose is to accommodate the dynamic and unpredictable needs of the internal cpu timing.
// For example, if tmr0 is set to roll over on a certain cycle and the program changes the
// pre-scale value, then the break point has to be moved to the new cycle.

bool Cycle_Counter::reassign_break(unsigned int old_cycle, unsigned int new_cycle, BreakCallBack *f=NULL)
{

  Cycle_Counter_breakpoint_list  *l1 = &active, *l2;

  bool found_old = 0;
  bool break_set = 0;

  //  cout << "Cycle_Counter::reassign_break, old " << old_cycle << " new " << new_cycle << '\n';

  //dump_breakpoints();
  while( (l1->next) && !found_old)
    {
      
      // If the next break point is at a cycle greater than the
      // one we wish to set, then we found the insertion point.
      // Otherwise 
	  if(l1->next->break_value ==  old_cycle)
	    {
	      //cout << " cycle match ";
	      if(l1->next->f == f)
		found_old = 1;
	    }
	  else
	    l1 = l1->next;

    }

  if(found_old)
    {
      // Now move the break point
      //cout << " found old ";

      if(new_cycle > old_cycle)
	{
	  // First check to see if we can stay in the same relative position within the list

	  // Is this the last one in the list? (or equivalently, is the one after this one a NULL)
	  if(l1->next->next == NULL)
	    {
	      //cout << " replaced at current position (next is NULL)\n";
	      l1->next->break_value = new_cycle;
	      break_on_this = active.next->break_value;
	      //dump_breakpoints();   // debug
	      return 1;
	    }

	  // Is the next one in the list still beyond this one?
	  if(l1->next->next->break_value > new_cycle)
	    {
	      //cout << " replaced at current position (next is greater)\n";
	      l1->next->break_value = new_cycle;
	      break_on_this = active.next->break_value;
	      //dump_breakpoints();   // debug
	      return 1;
	    }

	  // Darn. Looks like we have to move it.
	  //cout << " moving \n";

	  l2 = l1->next;                        // l2 now points to this break point

	  l1->next = l1->next->next;            // Unlink this break point

	  while( (l1->next) && !break_set)
	    {

	      // If the next break point is at a cycle greater than the
	      // one we wish to set, then we found the insertion point.
	      // Otherwise 
	      if(l1->next->break_value > new_cycle)
		break_set = 1;
	      else
		l1 = l1->next;

	    }

	  l2->next = l1->next;
	  l1->next = l2;

	  break_on_this = active.next->break_value;

	  l2->break_value = new_cycle;

	  //dump_breakpoints();   // debug
	}
      else      // old_cycle < new_cycle
	{
	  // First check to see if we can stay in the same relative position within the list
	  //cout << " old cycle is less than new one\n";
	  // Is this the first one in the list?
	  if(l1 == &active)
	    {
	      //cout << " replaced at current position\n";
	      l1->next->break_value = new_cycle;
	      break_on_this = new_cycle;
	      //dump_breakpoints();   // debug
	      return 1;
	    }

	  // Is the previous one in the list still before this one?
	  if(l1->break_value < new_cycle)
	    {
	      //cout << " replaced at current position\n";
	      l1->next->break_value = new_cycle;
	      //dump_breakpoints();   // debug
	      return 1;
	    }

	  // Darn. Looks like we have to move it.

	  l2 = l1->next;                        // l2 now points to this break point

	  l1->next = l1->next->next;            // Unlink this break point

	  l1 = &active;                         // Start searching from the beginning of the list

	  while( (l1->next) && !break_set)
	    {

	      // If the next break point is at a cycle greater than the
	      // one we wish to set, then we found the insertion point.
	      // Otherwise 
	      if(l1->next->break_value > new_cycle)
		break_set = 1;
	      else
		l1 = l1->next;

	    }

	  l2->next = l1->next;
	  l1->next = l2;

	  l2->break_value = new_cycle;

	  break_on_this = active.next->break_value;

	  //dump_breakpoints();   // debug
	}
    }
  else
    cout << " didn't find the old one\n";

}

void Cycle_Counter::clear_current_break(void)
{

  if(active.next == NULL)
    return;

  if(value == break_on_this)
    {
      // cout << "clearing current cycle break " << hex << setw(8) << setfill('0') << break_on_this <<'\n';
      Cycle_Counter_breakpoint_list  *l1;

      l1 = inactive.next;                  // ptr to 1st inactive bp
      inactive.next = active.next;         // let the 1st active bp become the 1st inactive one
      active.next = active.next->next;     // The 2nd active bp is now the 1st
      inactive.next->next = l1;            // The 2nd inactive bp used to be the 1st

      // %%%FIX ME%%% - there's a potential bug if the cycle counter rolls over and there's a breakpoint on cycle 0

      if(active.next != NULL)
	break_on_this = active.next->break_value;
      else
	break_on_this = 0;
    }
  else
    {
      // cout << "Didn't clear the current cycle break because != break_on_this\n";
    }
}

void Cycle_Counter::dump_breakpoints(void)
{
  Cycle_Counter_breakpoint_list  *l1 = &active;

  cout << "Next scheduled cycle break " << hex << setw(8) << setfill('0') << break_on_this << '\n';

  while(l1->next)
    {
      //cout << cpu->name_str << "  " << "internal cycle break  " <<
      cout << "internal cycle break  " <<
	hex << setw(8) << setfill('0') <<  l1->next->break_value << '\n';

      l1 = l1->next;
    }

}


Cycle_Counter::Cycle_Counter(void)
{
  value         = 0;
  break_on_this = 0;

  active.next = NULL;
  inactive.next = NULL;

  Cycle_Counter_breakpoint_list  *l1 = &inactive;

  for(int i=0; i<BREAK_ARRAY_SIZE; i++)
    {
      l1->next = new Cycle_Counter_breakpoint_list;
      l1 = l1->next;
    }
  l1->next = NULL;


}


//--------------------------------------------------
void WDT::put(unsigned int new_value)
{
  value = new_value;
}

void WDT::initialize(bool enable, double _timeout)
{
  break_point = 0;
  wdte = enable;
  timeout = _timeout;
  warned = 0;

  cout << " WDT init called. en = " << enable << " to = " << timeout << '\n';
  if(wdte)
    {
      cout << "Enabling WDT\n";

      value = cpu->time_to_cycles(timeout);
      prescale = cpu->option_reg.get_psa() ? (cpu->option_reg.get_prescale()) : 0;

      future_cycle = cpu->cycles.value + value * (1<<prescale);

      cpu->cycles.set_break(future_cycle, this);

    }

}

void WDT::callback(void)
{


  cout<<"WDT timeout: " << hex << cpu->cycles.value << '\n';

  prescale = cpu->option_reg.get_psa() ? (cpu->option_reg.get_prescale()) : 0;

  future_cycle = cpu->cycles.value + value * (1<<prescale);

  cpu->cycles.set_break(future_cycle, this);

  if(break_point)
    bp.halt();
  else
    cpu->reset(WDT_RESET);

}

void WDT::clear(void)
{
  if(wdte)
    {
      prescale = cpu->option_reg.get_psa() ? (cpu->option_reg.get_prescale()) : 0;

      unsigned int fc = cpu->cycles.value + value * (1<<prescale);

      // cout << "moving break from " << future_cycle << " to " << fc << '\n';

      cpu->cycles.reassign_break(future_cycle, fc, this);

      future_cycle = fc;
    }
  else
    {
      if(!warned)
	{
	  warned = 1;
	  cout << "The WDT is not enabled - clrwdt has no effect!\n";
	}
    }

}
//--------------------------------------------------
// WREG
//

unsigned int WREG::get(void)
{
  if(break_point) 
    bp.check_read_break(this);

  trace.read_W(value);

  return(value);
}

void WREG::put(unsigned int new_value)
{

  if(break_point) 
    bp.check_write_break(this);

  value = new_value;
  trace.write_W(value);

}

WREG::WREG(void)
{
  break_point = 0;
  new_name("W");
}


//--------------------------------------------------
// IOPORT
//
const int one_shifted_left_by_n [8] = 
{
  1 << 0,
  1 << 1,
  1 << 2,
  1 << 3,
  1 << 4,
  1 << 5,
  1 << 6,
  1 << 7
};

int IOPORT::update_stimuli(void)

{
  // Loop through the io pins and determine if there are
  // any sources attached to the same node

  //cout << "updating the stimuli\n";
  unsigned int time = cpu->cycles.value;
  int input = 0;

  for(int i = 0, m=1; i<IOPINS; i++, m <<= 1)
    if(stimulus_mask & m)
      {
	int t = pins[i]->snode->update(time);

	//cout << name() << ' ' << i;
	//cout << " pin " << pins[i]->name();
	//cout <<  " node "<< pins[i]->snode->name() << " is ";

	if(t  > pins[i]->threshold)
	  {
	    //cout << "above";
	     input |= m;
	  }
	//else cout << "below";

	//cout << " threshold. node " << t  << " threshold " << pins[i]->threshold << '\n';
      }
      
  return input;

}

int IOPORT::get_bit(unsigned int bit_number)
{

  unsigned int time = cpu->cycles.value;

  if(stimulus_mask & one_shifted_left_by_n [bit_number])
    return(pins[bit_number]->snode->update(time));
  else
    return( (value &  one_shifted_left_by_n [bit_number]) ? 1 : 0);

}

unsigned int IOPORT::get(void)
{

  //cout << "IOPORT::get()\n";

  // Update the stimuli - if there are any

  int input = 0;
  if(stimulus_mask)
    input = update_stimuli();   // %%% FIX ME %%% (portb pull up res can change states too)

  //  value = ( ( (value & ~stimulus_mask) | input) & tris->value) ;
  trace.register_read(address,value);

  return(value);
}

void IOPORT::put(unsigned int new_value)
{

  //if(break_point) 
  //  bp.check_write_break(this);

  // The I/O Ports have an internal latch that holds the state of the last
  // write, even if the I/O pins are configured as inputs. If the tris port
  // changes an I/O pin from an input to an output, then the contents of this
  // internal latch will be placed onto the external I/O pin.

  internal_latch = new_value;

  // update only those bits that are really outputs
  //cout << "IOPORT::put trying to put " << new_value << '\n';

  value = (new_value & ~tris->value) | (value & tris->value);

  //cout << " IOPORT::put just set port value to " << value << '\n';

  // Update the stimuli - if there are any
  if(stimulus_mask)
    update_stimuli();

  //cout << " IOPORT::put port value is " << value << " after updating stimuli\n";

  trace.register_write(address,value);

}


void IOPORT::setbit(unsigned int bit_number, bool new_value)
{

  int bit_mask = one_shifted_left_by_n[bit_number];
  //cout << name();

  if( ((bit_mask & value) != 0) ^ (new_value==1))
    {
      //cout << " IOPORT::set_bit bit changed due to a stimulus. new_value = " << new_value <<'\n';
      value ^= bit_mask;

      trace.register_write(address,value); // %%% FIX ME %%% give this another trace type.
    }
  //else cout <<  " IOPORT::set_bit bit did not change\n";

}

void IOPORT::attach_iopin(IOPIN * new_pin, unsigned int bit_position)
{
  //  cout << "trying to attach iopin\n";

  bit_position &= 7;

  pins[bit_position] = new_pin;

}

void IOPORT::attach_stimulus(stimulus *new_stimulus, unsigned int bit_position)
{
  //  cout << "trying to attach " << name() << " to " << new_stimulus->name() <<  '\n';

  bit_position &= 7;
  //cout << bit_position  << "  " << (1<<bit_position) << '\n';

  stimulus_mask |= (1<<bit_position);
  // io_stimulus[bit_position] = new_stimulus_node;

  if(pins[bit_position]->snode == NULL)
    {
      // If this I/O pin is not attached to a node yet, then create a node and attach it.
      pins[bit_position]->snode = new Stimulus_Node();
      pins[bit_position]->snode->attach_stimulus(pins[bit_position]);
    }

  // attach the new stimulus to the same node as this I/O pin's

  pins[bit_position]->snode->attach_stimulus(new_stimulus);

  //cout << "IOPORT attached new stimulus to " << pins[bit_position]->snode->name() << " named " <<
  //  pins[bit_position]->snode->name() << '\n';


}

void IOPORT::attach_node(Stimulus_Node *new_node, unsigned int bit_position)
{
  //  cout << "trying to attach " << new_node->name() << " to " << name() << " bit " << bit_position << '\n';

  if(pins[bit_position])
    {
      stimulus_mask |= (1<<bit_position);
      //      pins[bit_position]->snode == new_node;
    }
  else
    cout << "Error: attaching node to a non-existing I/O pin.\n";

}

void IOPORT::update_pin_directions(unsigned int new_tris)
{

  unsigned int diff = tris->value ^ new_tris;

  if(diff)
    {
      // First go through and update the direction of the I/O pins
      int i,m;
      for(i = 0, m=1; i<IOPINS; i++, m <<= 1)
	if(stimulus_mask & m & diff)
	    pins[i]->update_direction(m & (~new_tris));

      // Now, update the nodes to which the(se) pin(s) attach

      unsigned int time = cpu->cycles.value;
      for(i = 0, m=1; i<IOPINS; i++, m <<= 1)
	if(stimulus_mask & m & diff)
	  pins[i]->snode->update(time);
    }
}

IOPORT::IOPORT(void)
{
  break_point = 0;
  stimulus_mask = 0;

  tris = NULL;

  for(int i=0; i<IOPINS; i++)
    pins[i] = NULL;

  new_name("ioport");
}

//--------------------------------------------------
// IOPORT_TRIS
//

unsigned int IOPORT_TRIS::get(void)
{
  //  if(break_point) 
  //    bp.check_read_break(this);

  trace.register_read(address,value);

  return(value);
}

void IOPORT_TRIS::put(unsigned int new_value)
{

  //  if(break_point) 
  //    bp.check_write_break(this);
  port->update_pin_directions(new_value);

  value = new_value;

  // %%%FIX ME%%% - Should we update the I/O Port here too?

  trace.register_write(address,value);

}

IOPORT_TRIS::IOPORT_TRIS(void)
{
  break_point = 0;
  port = NULL;

  new_name("ioport");
}

//----------------------------------------------------------
//
// EE PROM
//
// There are many conditions that need to be verified against a real part:
//    1) what happens if RD and WR are set at the same time?
//       > the simulator ignores both the read and the write.
//    2) what happens if a RD is initiated while data is being written?
//       > the simulator ignores the read
//    3) what happens if EEADR or EEDATA are changed while data is being written?
//       > the simulator will update these registers with the new values that
//         are written to them, however the write operation will be unaffected.
//    4) if WRERR is set, will this prevent a valid write sequence from being initiated?
//       > the simulator views WRERR as a status bit
//    5) if a RD is attempted after the eeprom has been prepared for a write
//       will this affect the RD or write?
//       > The simulator will proceed with the read and leave the write-enable state alone.
//    6) what happens if WREN goes low while a write is happening?
//       > The simulator will complete the write and WREN will be cleared.

void EECON1::put(unsigned int new_value)
{

  if(break_point) 
    bp.check_write_break(this);

  if(new_value & WREN)
    {
      if(eeprom->eecon2->eestate == EECON2::UNARMED)
	{
	  eeprom->eecon2->eestate = EECON2::NOT_READY;
	  value |= WREN;

	}

      // WREN is true and EECON2 is armed (which means that we've passed through here
      // once before with WREN true). Initiate an eeprom write only if WR is true and
      // RD is false AND EECON2 is ready

      else if( (new_value & WR) && !(new_value & RD)  && (eeprom->eecon2->eestate == EECON2::READY))
	{
	  value |= WR;
	  eeprom->start_write();
	}

      //    else cout << "EECON1: write ignored " << new_value << "  (WREN is probably already set)\n";

    }
  else
    {
      // WREN is low so inhibit eeprom writes:

      eeprom->eecon2->eestate == EECON2::UNARMED;
      cout << "EECON1: write is disabled\n";

    }

  if ( (new_value & RD) && !( (new_value | value) & WR) )
    {
      eeprom->eedata->value = eeprom->rom[eeprom->eeadr->value]->get();
    }
  
  value = (new_value & (EEIF | WRERR | WREN)) | (value & ~(EEIF | WRERR | WREN)) ;

  trace.register_write(address,value);

}

unsigned int EECON1::get(void)
{
  if(break_point) 
    bp.check_read_break(this);

  trace.register_read(address,value);

  return(value);
}

EECON1::EECON1(void)
{
  new_name("eecon1");
  break_point = 0;
}





void EECON2::put(unsigned int new_value)
{

  if(break_point) 
    bp.check_write_break(this);


  if( (eestate == NOT_READY) && (0x55 == new_value))
    {
      eestate = HAVE_0x55;
    }
  else if ( (eestate == HAVE_0x55) && (0xaa == new_value))
    {
      eestate = READY;
    }
  else if ((eestate == HAVE_0x55) || (eestate == READY))
    {
      eestate == NOT_READY;
    }

  trace.register_write(address,new_value);

}

unsigned int EECON2::get(void)
{
  if(break_point) 
    bp.check_read_break(this);

  trace.register_read(address,0);

  return(0);
}

EECON2::EECON2(void)
{
  new_name("eecon2");
  break_point = 0;
}





unsigned int EEDATA::get(void)
{
  if(break_point) 
    bp.check_read_break(this);

  trace.register_read(address,value);

  return(value);
}

void EEDATA::put(unsigned int new_value)
{

  if(break_point) 
    bp.check_write_break(this);

  value = new_value;
  trace.register_write(address,value);

}

EEDATA::EEDATA(void)
{
  new_name("eedata");
  break_point = 0;
}




unsigned int EEADR::get(void)
{
  if(break_point) 
    bp.check_read_break(this);

  trace.register_read(address,value);

  return(value);
}

void EEADR::put(unsigned int new_value)
{

  if(break_point) 
    bp.check_write_break(this);

  value = new_value;
  trace.register_write(address,value);

}


EEADR::EEADR(void)
{
  new_name("eeadr");
  break_point = 0;
}



// EEPROM - Peripheral
//
//  This object emulates the 14-bit core's EEPROM/FLASH peripheral (such as the 16c84).
//
//  It's main purpose is to provide a means by which the control registers may communicate.
// 

EEPROM::EEPROM(void)
{

  rom_size = 0;
  eecon1 = NULL;
  eecon2 = NULL;
  eedata = NULL;
  eeadr  = NULL;


}


void EEPROM::start_write(void)
{
  cpu->cycles.set_break(cpu->cycles.value + EPROM_WRITE_TIME,this);

  wr_adr = eeadr->value;
  wr_data = eedata->value;

}

void EEPROM::callback(void)
{

  if(wr_adr < rom_size)
    rom[wr_adr]->value = wr_data;
  else
    cout << "EEPROM wr_adr is out of range " << wr_adr << '\n';

  unsigned int eecon1_value = eecon1->get();


  eecon1->value = (eecon1_value  & (~eecon1->WR)) | eecon1->EEIF;

  cpu->intcon->peripheral_interrupt();

  if (eecon1_value & eecon1->WREN)
    eecon2->eestate = EECON2::NOT_READY;
  else
    eecon2->eestate = EECON2::UNARMED;

}


void EEPROM::reset(RESET_TYPE by)
{

  switch(by)
    {
    case POR_RESET:
      eecon1->value = 0;          // eedata & eeadr are undefined at power up
      eecon2->eestate = EECON2::UNARMED;
    }

}

void EEPROM::initialize(unsigned int new_rom_size, EECON1 *con1, EECON2 *con2, EEDATA *data, EEADR *adr)
{

  rom_size = new_rom_size;

  // Save the pointers to all of the control registers
  
  eecon1 = con1;
  eecon2 = con2;
  eedata = data;
  eeadr  = adr;

  // Let the control registers have a pointer to the peripheral in which they belong.

  eecon1->eeprom = this;
  eecon2->eeprom = this;
  eedata->eeprom = this;
  eeadr->eeprom  = this;


  // Create the rom

  rom = (file_register **) new char[sizeof (file_register *) * rom_size];


  // Initialize the rom

  char str[100];
  for (int i = 0; i < rom_size; i++)
    {

      rom[i] = new file_register;
      rom[i]->address = i;
      rom[i]->break_point = 0;
      rom[i]->value = 0;

      sprintf (str, "eeprom reg 0x%02x", i);
      rom[i]->new_name(str);

    }

  reset(POR_RESET);

}

void EEPROM::dump(void)
{
  int i, j, reg_num,v;

  cout << "     " << hex;

  // Column labels
  for (i = 0; i < 16; i++)
    cout << setw(2) << setfill('0') <<  i << ' ';

  cout << '\n';

  for (i = 0; i < rom_size/16; i++)
    {
      cout << setw(2) << setfill('0') <<  i << ":  ";

      for (j = 0; j < 16; j++)
	{
	  reg_num = i * 16 + j;
	  if(reg_num < rom_size)
	    {
	      v = rom[reg_num]->value;
	      cout << setw(2) << setfill('0') <<  v << ' ';
	    }
	  else
	    cout << "-- ";
	}
      cout << "   ";

      for (j = 0; j < 16; j++)
	{
	  reg_num = i * 16 + j;
	  if(reg_num < rom_size)
	    {
	      v = rom[reg_num]->value;
	      if( (v >= ' ') && (v <= 'z'))
		cout.put(v);
	      else
		cout.put('.');
	    }
	}

      cout << '\n';

    }
}


