/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include <iostream.h>
#include <iomanip.h>

#include "pic-processor.h"
#include "breakpoints.h"
#include "14bit-processors.h"

extern unsigned int simulation_start_cycle;

Breakpoints bp;

unsigned int Breakpoints::set_breakpoint(BREAKPOINT_TYPES break_type, pic_processor *cpu,unsigned int arg1, unsigned arg2, BreakCallBack *f1 = NULL)
{
  file_register *fr;
  int i;
  Breakpoint_Instruction *abp;
  Break_register_read *bfrr;
  Break_register_write *bfrw;
  Break_register_read_value *bfrrv;
  Break_register_write_value *bfrwv;
  bool found =0;

  // First, find a free break point
  i=0;
  //  breakpoint_number = (last_breakpoint+ 1) % MAX_BREAKPOINTS;
  breakpoint_number = 0;

  do
    {
      if(break_status[breakpoint_number].type == BREAK_CLEAR)
	found=1;
      else
	breakpoint_number = (breakpoint_number + 1) % MAX_BREAKPOINTS;

      i++;
    } while((i < MAX_BREAKPOINTS) && !found);


  if(!found)
    {
      cout << "*** out of breakpoints\n";
      return(MAX_BREAKPOINTS);
    }
  //  else
  //  {
  //      last_breakpoint = breakpoint_number;
  //    }

  break_status[breakpoint_number].type = break_type;
  break_status[breakpoint_number].cpu  = cpu;
  break_status[breakpoint_number].arg1 = arg1;
  break_status[breakpoint_number].arg2 = arg2;
  break_status[breakpoint_number].f    = f1;


  switch (break_type)
    {
    case BREAK_ON_EXECUTION:

      // Get a new breakpoint instruction and let it replace
      // the instruction at which we are setting the break
      //   (note: the constructor saves the instruction we are replacing)

      if(arg1 < cpu->program_memory_size())
	{
	  cpu->program_memory[arg1] = 
	    new Breakpoint_Instruction(cpu,arg1,BREAK_ON_EXECUTION | breakpoint_number);

	  return(breakpoint_number);
	}
      else
	break_status[breakpoint_number].type = BREAK_CLEAR;

      break;

    case BREAK_ON_REG_WRITE:
      fr = cpu->registers[arg1];
      fr->break_point = BREAK_ON_REG_WRITE | breakpoint_number;
      bfrw = new Break_register_write();
      bfrw->replaced = fr;
      cpu->registers[arg1] = bfrw;
      return(breakpoint_number);
      break;

    case BREAK_ON_REG_WRITE_VALUE:
      fr = cpu->registers[arg1];
      fr->break_point = BREAK_ON_REG_WRITE_VALUE | breakpoint_number;
      bfrwv = new Break_register_write_value();
      bfrwv->replaced = fr;
      bfrwv->break_value = arg2;
      cpu->registers[arg1] = bfrwv;
      return(breakpoint_number);
      break;

    case BREAK_ON_REG_READ:
      fr = cpu->registers[arg1];
      fr->break_point = BREAK_ON_REG_READ | breakpoint_number;
      bfrr = new Break_register_read();
      bfrr->replaced = fr;
      cpu->registers[arg1] = bfrr;
      return(breakpoint_number);
      break;

    case BREAK_ON_REG_READ_VALUE:
      fr = cpu->registers[arg1];
      fr->break_point = BREAK_ON_REG_READ_VALUE | breakpoint_number;
      bfrrv = new Break_register_read_value();
      bfrrv->replaced = fr;
      bfrrv->break_value = arg2;
      cpu->registers[arg1] = bfrrv;
      return(breakpoint_number);
      break;

    case BREAK_ON_INVALID_FR:
      fr = cpu->registers[arg1];
      fr->break_point = BREAK_ON_INVALID_FR | breakpoint_number;
      return(breakpoint_number);
      break;

    case BREAK_ON_CYCLE:
      // The cycle counter does its own break points.
      if(cpu->cycles.set_break(arg1, f1, breakpoint_number))
	return(breakpoint_number);
      else
	break_status[breakpoint_number].type = BREAK_CLEAR;

      break;

    case BREAK_ON_WDT_TIMEOUT:
      ((_14bit_processor *)cpu)->wdt.break_point = BREAK_ON_WDT_TIMEOUT | breakpoint_number;
      return(breakpoint_number);
      break;

    default:   // Not a valid type
      break_status[breakpoint_number].type = BREAK_CLEAR;
      break;
    }

  return(MAX_BREAKPOINTS);
}


unsigned int  Breakpoints::set_execution_break(pic_processor *cpu, unsigned int address)
{
  return(set_breakpoint (Breakpoints::BREAK_ON_EXECUTION, cpu, address, 0));
}

unsigned int  Breakpoints::set_read_break(pic_processor *cpu, unsigned int register_number)
{
  return(set_breakpoint (Breakpoints::BREAK_ON_REG_READ, cpu, register_number, 0));
}

unsigned int  Breakpoints::set_write_break(pic_processor *cpu, unsigned int register_number)
{
  return(set_breakpoint (Breakpoints::BREAK_ON_REG_WRITE, cpu, register_number, 0));
}

unsigned int  Breakpoints::set_read_value_break(pic_processor *cpu, unsigned int register_number,unsigned int value)
{
  return(set_breakpoint (Breakpoints::BREAK_ON_REG_READ_VALUE, cpu, register_number, value));
}

unsigned int  Breakpoints::set_write_value_break(pic_processor *cpu, unsigned int register_number,unsigned int value)
{
  return(set_breakpoint (Breakpoints::BREAK_ON_REG_WRITE_VALUE, cpu, register_number, value));
}

unsigned int  Breakpoints::set_cycle_break(pic_processor *cpu, unsigned int future_cycle, BreakCallBack *f1 = NULL)
{

  return(set_breakpoint (Breakpoints::BREAK_ON_CYCLE, cpu, future_cycle, 0,f1));    
}

unsigned int  Breakpoints::set_wdt_break(pic_processor *cpu)
{
  // Set a wdt break only if one is not already set.

  if(cpu14->wdt.break_point == 0)
    return(set_breakpoint (Breakpoints::BREAK_ON_WDT_TIMEOUT, cpu, 0, 0));
  else
    return MAX_BREAKPOINTS;
}

//---------------------------------------------------------------------------------------
unsigned int Breakpoints::check_write_break(file_register *fr)
{
  cout << "debug   checking for write break point " << fr->address << '\n';

  if( (fr->break_point & BREAK_MASK) ==  BREAK_ON_REG_WRITE)
  {
    if(simulation_mode == RUNNING)
      {
	cout << "Hit a breakpoint!\n";
	trace.breakpoint( (Breakpoints::BREAK_ON_REG_WRITE>>8) );
	halt();
      }
    else
      {
	cout << "Stepped over a write breakpoint\n";
      }

    return(1);
  }

  return(0);
}

//---------------------------------------------------------------------------------------
unsigned int Breakpoints::check_read_break(file_register *fr)
{

  cout << "debug   checking for read break point " << fr->address << '\n';

  if( (fr->break_point & BREAK_MASK) ==  BREAK_ON_REG_READ)
  {
    if(simulation_mode == RUNNING)
      {
	cout << "Hit a breakpoint!\n";
	trace.breakpoint(fr->break_point & ~BREAK_MASK);
	halt();
      }
    else
      {
	cout << "Stepped over a read breakpoint\n";
      }

    return(1);
  }

  return(0);
}

//---------------------------------------------------------------------------------------
unsigned int Breakpoints::check_invalid_fr_break(invalid_file_register *fr)
{

  cout << "debug   checking for invalid file register access  break point " << fr->name() << '\n';

  if( (fr->break_point & BREAK_MASK) ==  BREAK_ON_INVALID_FR)
  {
    //global_break |= STOP_RUNNING;
    halt();
    trace.breakpoint(fr->break_point & ~BREAK_MASK);
    return(1);
  }

  return(0);
}

unsigned int Breakpoints::check_cycle_break(unsigned int abp)
{
  
  halt();
  if( abp < MAX_BREAKPOINTS)
    {
      if (NULL != break_status[abp].f)
	  break_status[abp].f->callback();

      trace.breakpoint( (Breakpoints::BREAK_ON_CYCLE>>8) );

      clear(abp);
    }

  return(1);

}
//---------------------------------------------------------------------------------------
unsigned int Breakpoints::check_break(file_register *fr)
{
  cout << "checking for a bp doesn't do anything...\n";

  return(0);
}

bool Breakpoints::dump1(unsigned int bp_num)
{

  bool set_by_user = 0;

  BREAKPOINT_TYPES break_type = break_status[bp_num].type;

  switch (break_type)
    {
    case BREAK_ON_EXECUTION:
      cout << hex << setw(0) << bp_num << ": " << break_status[bp_num].cpu->name_str << "  ";
      cout << "exec at " << hex << setw(4) << setfill('0') <<  break_status[bp_num].arg1 << '\n';
      set_by_user = 1;
      break;

    case BREAK_ON_REG_WRITE:
      cout << hex << setw(0) << bp_num << ": " << break_status[bp_num].cpu->name_str << "  ";
      cout << "reg write: " << hex << setw(2) << setfill('0') <<  break_status[bp_num].arg1 << '\n';
      set_by_user = 1;
      break;

    case BREAK_ON_REG_WRITE_VALUE:
      cout << hex << setw(0) << bp_num << ": " << break_status[bp_num].cpu->name_str << "  ";
      cout << "reg write. Break when " << hex << setw(2) << setfill('0') <<  break_status[bp_num].arg2 << 
	" is written to register " << break_status[bp_num].arg1 << '\n';
      set_by_user = 1;
      break;

    case BREAK_ON_REG_READ:
      cout << hex << setw(0) << bp_num << ": " << break_status[bp_num].cpu->name_str << "  ";
      cout << "reg read: " << hex << setw(2) << setfill('0') <<  break_status[bp_num].arg1 << '\n';
      set_by_user = 1;
      break;

    case BREAK_ON_CYCLE:
      cout << hex << setw(0) << bp_num << ": " << break_status[bp_num].cpu->name_str << "  ";
      cout << "cycle " << hex << setw(8) << setfill('0') <<  break_status[bp_num].arg1 << '\n';
      set_by_user = 1;
      break;

    case BREAK_ON_WDT_TIMEOUT:
      cout << hex << setw(0) << bp_num << ": " << break_status[bp_num].cpu->name_str << "  ";
      cout << "wdt time out\n";
      set_by_user = 1;
      break;

    }

  return(set_by_user);

}


void Breakpoints::dump(void)
{
  bool have_breakpoints = 0;


  for(int i = 0; i<MAX_BREAKPOINTS; i++)
    {

      if(dump1(i))
	have_breakpoints = 1;
    }


  if(!have_breakpoints)
    cout << "No user breakpoints are set... I think\n";

}


void Breakpoints::clear(unsigned int b)
{
  Breakpoint_Instruction *abp;
  instruction *inst;
  file_register *fr;
  Break_register_read *bfrr;
  Break_register_write *bfrw;
  Break_register_read_value *bfrrv;
  Break_register_write_value *bfrwv;

  if(b<MAX_BREAKPOINTS)
    {
      BreakStatus bs = break_status[b];   // 

      switch (bs.type)
	{
	case BREAK_ON_EXECUTION:

	  inst = bs.cpu->program_memory[bs.arg1];              // Get a pointer to the break point instruction
	  abp= (Breakpoint_Instruction *) inst;
	  bs.cpu->program_memory[bs.arg1] = abp->replaced;     // Restore the instruction

	  delete abp;
	  break_status[b].type = BREAK_CLEAR;
	  cout << "Cleared execution breakpoint number " << b << '\n';
	  //if(b<last_breakpoint)
	  //  last_breakpoint = b-1;

	  break;

	case BREAK_ON_CYCLE:
	  break_status[b].type = BREAK_CLEAR;
	  cout << "Cleared cycle breakpoint number " << b << '\n';
	  //if(b<last_breakpoint)
	  //  last_breakpoint = b-1;

	  break;

	case BREAK_ON_REG_READ:
	  fr = bs.cpu->registers[bs.arg1];
	  if (fr->isa() == file_register::BP_REGISTER )
	    {
	      bfrr = (Break_register_read *)fr;
	      bs.cpu->registers[bs.arg1] = bfrr->replaced;
	      delete bfrr;
	      break_status[b].type = BREAK_CLEAR;
	      cout << "Cleared breakpoint number " << b << " on register " << bs.arg1 << '\n';
	      //if(b<last_breakpoint)
	      //last_breakpoint = b-1;

	    }
	  else
	    {
	      cout << "??? break is not set\n";
	    }

	  break;

	case BREAK_ON_REG_READ_VALUE:
	  fr = bs.cpu->registers[bs.arg1];
	  if (fr->isa() == file_register::BP_REGISTER )
	    {
	      bfrrv = (Break_register_read_value *)fr;
	      bs.cpu->registers[bs.arg1] = bfrrv->replaced;
	      delete bfrrv;
	      break_status[b].type = BREAK_CLEAR;
	      cout << "Cleared breakpoint number " << b << " on register " << bs.arg1 << '\n';
	      //if(b<last_breakpoint)
	      //last_breakpoint = b-1;

	    }
	  else
	    {
	      cout << "??? break is not set\n";
	    }

	  break;

	case BREAK_ON_REG_WRITE:
	  fr = bs.cpu->registers[bs.arg1];
	  if (fr->isa() == file_register::BP_REGISTER )
	    {
	      bfrw = (Break_register_write *)fr;
	      bs.cpu->registers[bs.arg1] = bfrw->replaced;
	      delete bfrw;
	      break_status[b].type = BREAK_CLEAR;
	      cout << "Cleared breakpoint number " << b << " on register " << bs.arg1 << '\n';
	      //if(b<last_breakpoint)
	      //last_breakpoint = b-1;

	    }
	  else
	    {
	      cout << "??? break is not set\n";
	    }

	  break;

	case BREAK_ON_REG_WRITE_VALUE:
	  fr = bs.cpu->registers[bs.arg1];
	  if (fr->isa() == file_register::BP_REGISTER )
	    {
	      bfrwv = (Break_register_write_value *)fr;
	      bs.cpu->registers[bs.arg1] = bfrwv->replaced;
	      delete bfrwv;
	      break_status[b].type = BREAK_CLEAR;
	      cout << "Cleared breakpoint number " << b << " on register " << bs.arg1 << '\n';
	      //if(b<last_breakpoint)
	      //last_breakpoint = b-1;

	    }
	  else
	    {
	      cout << "??? break is not set\n";
	    }

	  break;


	case BREAK_ON_WDT_TIMEOUT:
	  bs.type = BREAK_CLEAR;
	  cout << "Cleared wdt timeout breakpoint number " << b << '\n';
	  ((_14bit_processor *)bs.cpu)->wdt.break_point = 0;
	  //if(b<last_breakpoint)
	  //last_breakpoint = b-1;

	  break;

	}
    }
}

//
//  dump_traced
//  Called by the trace class to display a breakpoint that is in the
// trace buffer.

void Breakpoints::dump_traced(unsigned int b)
{

  BREAKPOINT_TYPES break_type = (BREAKPOINT_TYPES) ((b & 0xff0000) << 8);

  switch (break_type)
    {
    case BREAK_ON_EXECUTION:
      cout << "execution at "<< hex << setw(4) << setfill('0') <<  (b & 0xffff) << '\n';
      break;

    case BREAK_ON_REG_WRITE:
      cout << "reg write: " << hex << setw(2) << setfill('0') <<  (b & 0xff) << '\n';
      break;

    case BREAK_ON_REG_WRITE_VALUE:
      cout << "wrote " << hex << setw(2) << setfill('0') <<  ((b & 0xff00)>>8) << 
	" to register " << hex << setw(2) << setfill('0') <<  (b & 0xff) << '\n';
      break;

    case BREAK_ON_REG_READ:
      cout << "reg write: " << hex << setw(2) << setfill('0') <<  (b & 0xff) << '\n';
      break;

    case BREAK_ON_CYCLE:
      cout << "cycle " << '\n';
      break;

    case BREAK_ON_WDT_TIMEOUT:
      cout << "wdt time out\n";
      break;

    default:
      cout << "unknown\n";
    }



}


// Clear all break points that are set for a specific processor
// This only be called when a processor is being removed and not when a user 
// wants to clear the break points. Otherwise, internal break points like
// invalid register accesses will get cleared.

void Breakpoints::clear_all(pic_processor *c)
{

  for(int i=0; i<MAX_BREAKPOINTS; i++)
    {
      if(c == break_status[i].cpu)
	clear(i);
    }

}

void Breakpoints::clear_all_set_by_user(pic_processor *c)
{

  for(int i=0; i<MAX_BREAKPOINTS; i++)
    {
      if((c == break_status[i].cpu) && (break_status[i].type != BREAK_ON_INVALID_FR))
	clear(i);
    }

}


Breakpoints::Breakpoints(void)
{
  
  //  execution_breakpoint_ptr = new Breakpoint_Instruction(memory_size);
  breakpoint_number = 0;
  //last_breakpoint =0;

  for(int i=0; i<MAX_BREAKPOINTS; i++)
    break_status[i].type = BREAK_CLEAR;

}

//---------------------------------------------------------------------------------------
void Breakpoint_Instruction::execute(void)
{

  if( (simulation_mode == RUNNING) && (simulation_start_cycle != cpu->cycles.value))
    {
      cout << "Hit a breakpoint!\n";
      trace.breakpoint( (Breakpoints::BREAK_ON_EXECUTION>>8) | address );
      bp.halt();
    }
  else
    {
      cout << "Ignoring a breakpoint\n";
      replaced->execute();
    }

}

Breakpoint_Instruction::Breakpoint_Instruction(pic_processor *new_cpu, unsigned int new_address,unsigned int bp)
{
  cpu = new_cpu;
  address = new_address;
  opcode = 0xffffffff;
  bpn = bp;
  replaced = cpu->program_memory[address];
}

int Breakpoint_Instruction::get_src_line(void)
{
  return(replaced->get_src_line());
}
int Breakpoint_Instruction::get_lst_line(void)
{
  return(replaced->get_lst_line());
}

char * Breakpoint_Instruction::name(char *return_str)
{

  return(replaced->name(return_str));
}

char * Break_register_read_value::name(void)
{
  if(replaced)
    return(replaced->name());
}

char * Break_register_write_value::name(void)
{
  if(replaced)
    return(replaced->name());
}

char * Break_register_read::name(void)
{
  if(replaced)
    return(replaced->name());
}

char * Break_register_write::name(void)
{
  if(replaced)
    return(replaced->name());
}

//====================================================================================
//
// catch_control_c
//
//  
void catch_control_c(int sig)
{

  if(simulation_mode != STOPPED)
    {
      cout << "<CTRL C> break\n";
      bp.halt();
    }
  else
    cout << "caught control c, but it doesn't seem gpsim was simulating\n";


}

