/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


/* input.c -- adapted from fileman.c - a simple program illustrating how to use
   the GNU Readline library.
   */

#include <stdio.h>
extern "C" {
#include <readline/readline.h>
}

extern "C" {
#include <readline/history.h>
}

#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <g++/string>
#include <g++/list>
#include <g++/csignal>

#include "commands.h"

#define MAX_LINE_LENGTH 256  


//========================================================================================
// A lot of the crap in here is a carry over from the example fileman.c in the readline
// library. Expect it to disappear...

/* The names of functions that actually do the manipulation. */
int com_list (char *);
int com_view (char *);
/*int com_pwd (char *);*/
int com_help (char *); //, com_cd (char *); //, com_quit (char *);

//
void initialize_readline (void);
extern void open_stimulus_file(FILE *);


/* When non-zero, this global means the user is done using this program. */
int done = 0;

const char *gpsim = "gpsim> ";  // Normal prompt
const char *gpsim_cont = "> ";  // command continuation prompt

COMMAND_MODES command_mode;


//====================================================================================
extern void catch_control_c(int); // In breakpoints.cc

void initialize_signals(void)
{

  static struct sigaction action;

  action.sa_handler = catch_control_c;
  sigemptyset(&action.sa_mask);
  action.sa_flags=0;

  sigaction(SIGINT, &action, 0);


}
//====================================================================================
//
// bool parse_line
//
//  This c function will parse a string of characters into it's constituent substrings
// in a way somewhat analogous to getopt. In other words, if you imagine the string
// as a sentence, parse_line would create the argc/argv pair that would be filled
// with the words.
//  A space (' ') and a tab ('\t') are used as the word delimeters while an octothorpe
// (don't invite me to a party) '#' is used as the comment character (i.e. everything
// after the comment character gets ignored.


bool parse_line(char *line)
{

#define COMMENT_CHAR '#'

  register int i;

  bool searching_for_whitespace;
  bool comment = 0;
  char *argv_ptr=NULL;

  // parse the line in the same way the standard C program is invoked
  // i.e. create an argc / argv pair:

  searching_for_whitespace = 1;          // Start searching for white space
  i = 0;
  argc = 0;
  *argv[0] = '#';                        // pre-condition a comment

  while(line[i] && !comment)
    {
      if( !((line[i] == ' ') || (line[i] == '\t')) )  // white space ?
	{
	  if(line[i] != COMMENT_CHAR)
	    {
	      if(searching_for_whitespace)            // If this is the first non-white space
		{
		  searching_for_whitespace = 0;       // then indicate that we're getting it
		  argv_ptr = argv[argc];              // and get the ptr to where we'll place it
		}
	      if(line[i] != '\n')
		*argv_ptr++ = line[i];                //
	    }
	  else
	    {
	    comment = 1;
	    if((searching_for_whitespace) && (argc>0))
	      argc--;
	    }
	}
      else if(searching_for_whitespace==0)            // Found first white space after non-white space
	{
	  *argv_ptr = '\0';                           // terminate the string.
	  searching_for_whitespace = 1;               // and then indicate we're searching for white space
	  argc++;
	}
      i++;                                            //
    }

  if(argv_ptr)
    {
      if(*argv[0] == COMMENT_CHAR )  // is this whole line a comment ?
	return 0;

      *argv_ptr = '\0';

      return 1;
    }
  return 0;
}

/* Execute a command line. */
COMMAND_MODES execute_line (char *line)
{

  static Command *a_command=NULL;
  static COMMAND_MODES command_mode = NO_COMMAND;
  string w = argv[0];

  if(!parse_line(line))
    return command_mode;

  if(command_mode == NO_COMMAND)
    {
      /* Isolate the command word. */
      w = argv[0];

      a_command = find_command (w);

      if (!a_command)
	{
	  fprintf (stderr, "%s: No such command for gpsim.\n", argv[0]);
	  return command_mode;
	}
    }

  /* Call the function. */

  command_mode = a_command->execute(argc,argv);
  return command_mode;
}

//--------------------
//
//

void process_command_file(char * file_name)
{

  char ss[MAX_LINE_LENGTH];
  FILE *f = fopen(file_name,"r");

  if(f)
    {
      while( fgets(ss, MAX_LINE_LENGTH, f) != NULL)
	execute_line(ss);

      fclose(f);
    }

}

void initialize_input(void)
{

  for(int i =0; i<MAX_ARGUMENTS; i++)
    argv[i] = new char[MAX_LINE_LENGTH];

}

main_input (void)
{

  COMMAND_MODES command_mode=NO_COMMAND;

  // if there is a startup file then read it and do what it says
  //  if(startup)
  //  process_command_file(startup);

  initialize_readline ();	/* Bind our completer. */

  initialize_signals();

  char last_line[256];
  last_line[0] = 0;

  int flag = 0;

  /* Loop reading and executing lines until the user quits. */
  while (!quit_gpsim)
    {
      char *line;
      
      if(command_mode == NO_COMMAND)
	{
	  line = ::readline (gpsim);
	}
      else
	{
	  line = ::readline (gpsim_cont);
	}
      if(flag) {
	execute_line( "load c analog_stim.stc\n");
	flag =0;
      }

      if (!line)
	{
	  done = 1;		/* Encountered EOF at top level. */
	}
      else
	{
	  /* Remove leading and trailing whitespace from the line.
	     Then, if there is anything left, add it to the history list
	     and execute it. */
	  stripwhite (line);

	  if (*line)
	    {
	      add_history (line);
	      command_mode = execute_line (line);
	      strncpy(last_line,line,256);
	    }
	  else
	    {
	      cout << "empty line\n";
	      if(*last_line)
		command_mode = execute_line (last_line);
	    }
	}

      if (line)
	free (line);
      line = (char *)NULL;

    }
  exit (0);
}

/* **************************************************************** */
/*                                                                  */
/*                  Interface to Readline Completion                */
/*                                                                  */
/* **************************************************************** */

/* Tell the GNU Readline library how to complete.  We want to try to complete
   on command names if this is the first word in the line, or on filenames
   if not. */
void initialize_readline (void)
{
  char **gpsim_completion (char *, int, int);

  /* Allow conditional parsing of the ~/.inputrc file. */
  rl_readline_name = "gpsim";

  /* Tell the completer that we want a crack first. */
  rl_attempted_completion_function = (CPPFunction *)gpsim_completion;
}

/* Attempt to complete on the contents of TEXT.  START and END show the
   region of TEXT that contains the word to complete.  We can use the
   entire line in case we want to do some simple parsing.  Return the
   array of matches, or NULL if there aren't any. */
char **
gpsim_completion (char *text, int start, int end)
{
  char **matches;
  char *command_generator (char *, int);


  matches = (char **)NULL;

  /* If this word is at the start of the line, then it is a command
     to complete.  Otherwise it is the name of a file in the current
     directory. */
  if (start == 0)
    matches = completion_matches (text, command_generator);

  return (matches);
}

/* Generator function for command completion.  STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the list. */
char *
command_generator (char *text, int state)
{
  static CL_iterator cli;
  char  *n;

  /* If this is a new word to complete, initialize now.  */

  if (!state)
    cli = command_list.begin();


  /* Return the next name which partially matches from the command list. */
  while (cli != command_list.end())
    {
      Command *t = *cli++;

      if(t->name.find(text) == 0)
	{
	  n = (char *) malloc(t->name.size() * sizeof(char));
	  if(n == NULL)
	    {
	      fprintf (stderr, "malloc: Out of virtual memory!\n");
	      abort ();
	    }
	  strcpy(n,t->name.c_str());
	  return(n);
	}

    }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

/* **************************************************************** */
/*                                                                  */
/*             gpsim Commands stolen from fileman.c                 */
/*                   (these are not used right now)                 */
/* **************************************************************** */

/* Return non-zero if ARG is a valid argument for CALLER, else print
   an error message and return zero. */
int
valid_argument (char *caller, char *arg)
{
  if (!arg || !*arg)
    {
      fprintf (stderr, "%s: Argument required.\n", caller);
      return (0);
    }

  return (1);
}

/* String to pass to system ().  This is for the LIST, VIEW and RENAME
   commands. */
static char syscom[1024];

/* List the file(s) named in arg. */
com_list (char *arg)
{
  if (!arg)
    arg = "*";

#ifdef __GO32__
  sprintf (syscom, "ls -lp %s", arg);
#else
  sprintf (syscom, "ls -FClg %s", arg);
#endif
  system (syscom);
}

com_view (char *arg)
{
  if (!valid_argument ("view", arg))
    return 0;

  sprintf (syscom, "cat %s | more", arg);
  system (syscom);
}


/* Print out the current working directory. */
/*
com_pwd (char *ignore)
{
  char dir[1024];

  (void) getwd (dir);

  printf ("Current directory is %s\n", dir);
}
*/

/* Change to the directory ARG. */
/*
com_cd (char *arg)
{
  if (chdir (arg) == -1)
    perror (arg);

  com_pwd ("");
}
*/
