/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */



// Portions of this files are (C) by Ian King:

/* picdis.c  - pic disassembler         */
/* version 0.1                          */
/* (c) I.King 1994                      */

#include <stdio.h>
#include <iostream.h>
#include <iomanip.h>
#include <g++/string>
#include <g++/list>

#include "commands.h"
#include "gpsim_def.h"


void welcome (char *);

unsigned int config_word;  // %%%FIX_ME%%%

extern void initialize_input(void);

void 
welcome (char *iam)
{
  printf ("\n\nuseage:\n%s [-h] [-p<device> [<hex_file>]] [-c<stc_file>]\n", iam);
  printf ("\t-h            : this help list\n");
  printf ("\t-p<device>    : processor (e.g. -pp16c84 for the 'c84)\n");
  printf ("\t<hex_file>    : input file in \"intelhex16\" format\n");
  printf ("\t-c<stc_file>  : startup command file\n");
  printf ("\t-s<cod_file>  : .cod symbol file\n\n");
  printf ("\t-v            : gpsim version\n");
}





void 
main (int argc, char *argv[])
{

  FILE *inputfile = stdin, *startup=NULL;
#define FILE_STRING_LENGTH 50
  char 
    startup_name[FILE_STRING_LENGTH] = "",
    pic_name[FILE_STRING_LENGTH]   = "",
    cod_name[FILE_STRING_LENGTH]   = "",
    hex_name[FILE_STRING_LENGTH]   = "";

  int i;
  int j;
  char command_str[256];
  int architecture;


  initialize_commands(NULL);
  initialize_input();


  cout << hex ;
  cin >> hex;

  i = 1;

  while(i < argc)
    {
      if (argv[i][0] == '-')
	switch (argv[i][1])
	  {
	  case 'P':
	  case 'p':
	    strncpy(pic_name, &argv[i][2],FILE_STRING_LENGTH);
	    break;

          case 'C':
          case 'c':
	    printf("%s  %s\n", &argv[i][0], &argv[i][2]);
	    strncpy (startup_name, &argv[i][2],FILE_STRING_LENGTH);
	    break;

          case 'S':
          case 's':
	    printf("%s  %s\n", &argv[i][0], &argv[i][2]);
	    strncpy (cod_name, &argv[i][2],FILE_STRING_LENGTH);
	    break;

	  case 'V':
	  case 'v':
	    printf("%s\n",GPSIM_VERSION);
	    break;

	  case 'H':
          case 'h':
	  default:
	    welcome (argv[0]);
	    exit (1);

	  }
      else
	{
	  strncpy(hex_name, argv[i],FILE_STRING_LENGTH);
	}

      i++;
    }


  if(*pic_name)
    {
      strcpy(command_str, "processor ");
      strcat(command_str, pic_name);
      execute_line(command_str);
    }

  if(*hex_name)
    {
      strcpy(command_str, "load h ");
      strcat(command_str, hex_name);
      execute_line(command_str);
    }

  if(*cod_name)
    {
      strcpy(command_str, "load s ");
      strcat(command_str, cod_name);
      execute_line(command_str);
    }

  if(*startup_name)
    {
      strcpy(command_str, "load c ");
      strcat(command_str, startup_name);
      execute_line(command_str);
    }



  main_input();
}


/* ... The End ... */
