/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#ifndef __P16X6X_H__
#define __P16X6X_H__

#include "14bit-processors.h"
#include "14bit-tmrs.h"


class PORTC : public IOPORT
{
public:

  /* Define the I/O pins that are common among the various PIC's
   * endowed with a port C.
   */

  const int
    T1CKI = 1 << 0,
    CCP1  = 1 << 2,
    SCK   = 1 << 3,
    SCL   = 1 << 3,  /* SCL and SCK share the same pin */
    SDI   = 1 << 4,
    SDA   = 1 << 4,  /* SDA and SDI share the same pin */
    SDO   = 1 << 5;

  /* Now define the I/O pins that are device dependent */
  const int
    T1OSO = 1 << 0,
    T1OSI = 1 << 1,
    _T1OSO = 1 << 1,  /* For some bizarre reason, TMR1's OS interface */
    _T1OSI = 1 << 0,  /* varies from pic to pic... */
    CCP2  = 1 << 1,
    TX    = 1 << 6,  /* Not all pics with a port C have a usart, */
    CK    = 1 << 6,  /* but the ones that do share TX and CK */
    RX    = 1 << 7,  /* Same goes for RX and DT */
    DT    = 1 << 7;

  /* A flag to sort out how the TMR1 OS interface is implemented.
   *   set: T1OSO == RC0 && T1OSI == RC1
   * clear: T1OSO == RC1 && T1OSI == RC0
   */
  bool t1oso_order;

  CCPCON *ccp1con;

  unsigned int get(void);
  void setbit(unsigned int bit_number, bool new_value);
};


class INTCON_P16C6x : public INTCON
{
public:

  PIR1 *pir1;
  PIR2 *pir2;

  bool check_peripheral_interrupt(void)
    {
      if(pir2)
	return( pir1->interrupt_status() | pir2->interrupt_status());
      else
	return( pir1->interrupt_status());
    }
 INTCON_P16C6x(void)
   {
     pir2=NULL;
   }
};

class P16C61 : public  _14bit_processor, public _14bit_18pins
{
public:

  const int PROGRAM_MEMORY_SIZE = 0x400;

  virtual PROCESSOR_TYPE isa(void){return _P16C61_;};
  virtual void create_symbols(void);

  virtual unsigned int program_memory_size(void) { return PROGRAM_MEMORY_SIZE; };
  virtual const GPR_map * get_GPR_map(void) { return NULL;};
  virtual int GPR_map_size(void) { return 0;};

  virtual void create_sfr_map(void);

  virtual void option_new_bits_6_7(unsigned int bits)
    {
      portb.rbpu_intedg_update(bits);
    }

  P16C61(void);

};

class _14bit_40pins
{
public:


  PORTA        porta;
  IOPORT_TRIS  trisa;

  PORTB        portb;
  IOPORT_TRIS  trisb;

  PORTC        portc;
  IOPORT_TRIS  trisc;

  /*
  IOPORT       portd;
  IOPORT_TRIS  trisd;
  */

  T1CON   t1con;
  PIR1    pir1;
  PIE     pie1;
  PIR2    pir2;
  PIE     pie2;
  T2CON   t2con;
  PR2     pr2;
  TMR2    tmr2;
  TMR1L   tmr1l;
  TMR1H   tmr1h;
  CCPCON  ccp1con;
  CCPRL   ccpr1l;
  CCPRH   ccpr1h;
  CCPCON  ccp2con;
  CCPRL   ccpr2l;
  CCPRH   ccpr2h;
  INTCON_P16C6x intcon_reg;

  void create_iopin_map(IOPIN_map ** iopin_map_ptr, int * num_of_iopins_ptr);

};

class P16C64 : public  _14bit_processor, public _14bit_40pins
{
  public:


  const int PROGRAM_MEMORY_SIZE = 0x800;

  TMR2_MODULE tmr2_module;

  virtual PROCESSOR_TYPE isa(void){return _P16C64_;};
  virtual void create_symbols(void);

  virtual unsigned int program_memory_size(void) { return PROGRAM_MEMORY_SIZE; };
  virtual const GPR_map * get_GPR_map(void) { return NULL;};
  virtual int GPR_map_size(void) { return 0;};

  virtual void create_sfr_map(void);

  virtual void option_new_bits_6_7(unsigned int bits)
    {
      portb.rbpu_intedg_update(bits);
    }

  P16C64(void);


};

class P16C65 : public  P16C64
{
  public:


  const int PROGRAM_MEMORY_SIZE = 0x800;


  virtual PROCESSOR_TYPE isa(void){return _P16C65_;};
  virtual void create_symbols(void);

  virtual unsigned int program_memory_size(void) { return PROGRAM_MEMORY_SIZE; };

  virtual void create_sfr_map(void);


  P16C65(void);


};


#endif
