/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#ifndef __P16X8X_H__
#define __P16X8X_H__

#include "14bit-processors.h"


class P16C8x : public  _14bit_processor, public _14bit_18pins
{
public:

  virtual PROCESSOR_TYPE isa(void){return _P16C84_;};
  virtual void create_symbols(void);

  virtual unsigned int program_memory_size(void) { return 0; };
  virtual const GPR_map * get_GPR_map(void) { return NULL;};
  virtual int GPR_map_size(void) { return 0;};
  virtual void create_sfr_map(void);
  virtual void option_new_bits_6_7(unsigned int bits)
    {
      portb.rbpu_intedg_update(bits);
    }


};

class P16C84 : public P16C8x
{
public:

  const int PROGRAM_MEMORY_SIZE = 0x400;

  static const GPR_map gpr_map[] =
  {
    {0x0c, 0x2f, 0x80}
  };
  const int num_of_gprs = sizeof (gpr_map) / sizeof (GPR_map);


  virtual PROCESSOR_TYPE isa(void){return _P16C84_;};

  virtual unsigned int program_memory_size(void) { return PROGRAM_MEMORY_SIZE; };
  virtual const GPR_map * get_GPR_map(void) { return gpr_map;};
  virtual int GPR_map_size(void) { return ( sizeof(gpr_map) / sizeof (GPR_map));};

  P16C84(void);

};

class P16F84 : public P16C8x
{
public:

  const int PROGRAM_MEMORY_SIZE = 0x400;

  static const GPR_map gpr_map[] =
  {
    {0x0c, 0x4f, 0x80}
  };
  const int num_of_gprs = sizeof (gpr_map) / sizeof (GPR_map);


  virtual PROCESSOR_TYPE isa(void){return _P16F84_;};

  virtual unsigned int program_memory_size(void) { return PROGRAM_MEMORY_SIZE; };
  virtual const GPR_map * get_GPR_map(void) { return gpr_map;};
  virtual int GPR_map_size(void) { return ( sizeof(gpr_map) / sizeof (GPR_map));};


  P16F84(void);

};

class P16CR84 : public P16F84
{
public:

  virtual PROCESSOR_TYPE isa(void){return _P16CR84_;};

  P16CR84(void) {  name_str = "p16cr84"; };

};



class P16F83 : public P16C8x
{
public:

  const int PROGRAM_MEMORY_SIZE = 0x200;

  static const GPR_map gpr_map[] =
  {
    {0x0c, 0x2f, 0x80}
  };
  const int num_of_gprs = sizeof (gpr_map) / sizeof (GPR_map);


  virtual PROCESSOR_TYPE isa(void){return _P16F83_;};

  virtual unsigned int program_memory_size(void) { return PROGRAM_MEMORY_SIZE; };
  virtual const GPR_map * get_GPR_map(void) { return gpr_map;};
  virtual int GPR_map_size(void) { return ( sizeof(gpr_map) / sizeof (GPR_map));};


  P16F83(void);

};

class P16CR83 : public P16F83
{
public:

  virtual PROCESSOR_TYPE isa(void){return _P16CR83_;};

  P16CR83(void) {  name_str = "p16cr83"; };

};


#endif
